// Copyright Notice
// ================
// BOCHS is Copyright 1994-1998 by Kevin P. Lawton.
//
// This file (SB16.CC) Copyright 1998-1999 by Josef Drexler
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
// 
//     EMail:    bochs@world.std.com




#include "bochs.h"


#if BX_USE_SB16_SMF
bx_sb16_c bx_sb16;
#define this NULL
#endif

bx_sb16_c::bx_sb16_c(void)
{
  // nothing for now
}

bx_sb16_c::~bx_sb16_c(void)
{
  if (bx_options.sb16.loglevel > 0)
    fclose(LOGFILE);

  if (bx_options.sb16.midimode == 2)
    finishmidifile();
  else if (bx_options.sb16.midimode == 1)
    closemidioutput();

  if (bx_options.sb16.wavemode == 2)
    finishvocfile();
  else if (bx_options.sb16.wavemode == 1)
    closewaveoutput();
}

  void
bx_sb16_c::init(bx_devices_c *d)
{
  BX_SB16_THIS devices = d;

  if ( (bx_options.sb16.logfile == NULL) ||
       (strlen(bx_options.sb16.logfile) < 1) )
    bx_options.sb16.loglevel = 0;

  if (bx_options.sb16.loglevel > 0)
    LOGFILE = fopen(bx_options.sb16.logfile,"w"); // logfile for errors etc.

  if (bx_options.sb16.midimode == 2)
    {
      MIDIDATA = fopen(bx_options.sb16.midifile,"w");
      if (MIDIDATA == NULL)
	{
	  writelog (MIDILOG(2), "Error opening file %s: %s. Midimode disabled.",
		    bx_options.sb16.midifile, strerror(errno));
	  bx_options.sb16.midimode = 0;
	}
      else
	initmidifile();
    }
  else if (bx_options.sb16.midimode == 1)
    openmidioutput();

  if (bx_options.sb16.wavemode == 2)
    {
      WAVEDATA = fopen(bx_options.sb16.wavefile,"w");
      if (WAVEDATA == NULL)
	{
	  writelog (WAVELOG(2), "Error opening file %s: %s. Wavemode disabled.",
		    bx_options.sb16.wavefile, strerror(errno));
	  bx_options.sb16.wavemode = 0;
	}
      else
	initvocfile();
    }
  else if (bx_options.sb16.wavemode == 1)
    openwaveoutput();

  fprintf(stderr, "SB16 options: midi=%d,%s  wave=%d,%s  log=%d,%s\n",
	  bx_options.sb16.midimode, bx_options.sb16.midifile,
	  bx_options.sb16.wavemode, bx_options.sb16.wavefile,
	  bx_options.sb16.loglevel, bx_options.sb16.logfile);

  // allocate the FIFO buffers - except for the MPUMIDICMD buffer
  // these sizes are generous, 16 or 8 would probably be sufficient
  MPU.datain.init( (int) 64);	        // the input
  MPU.dataout.init( (int) 64);	        // and output
  MPU.cmd.init( (int) 64);              // and command buffers
  MPU.midicmd.init( (int) 256);         // and the midi command buffer
  DSP.datain.init( (int) 64);           // the DSP input
  DSP.dataout.init( (int) 64);          // and output buffers
  EMUL.datain.init( (int) 64);          // the emulator ports
  EMUL.dataout.init( (int) 64);         // for changing emulator settings


  // reset all parts of the hardware by
  // triggering their reset functions

  // reset the Emulator port
  emul_write(0x00);

  // reset the MPU401
  mpu_command(0xff);

  // reset the DSP
  DSP.resetport = 1;  // so that one call to dsp_reset is sufficient
  dsp_reset(0);       // (reset is 1 to 0 transition)

  BX_SB16_IRQ = -1; // will be initialized later by the mixer reset

  // call the mixer reset 
  mixer_writeregister(0x00);
  mixer_writedata(0x00);

// Allocate the IO addresses, 2x0..2xf and 3x0..3x4
  for (unsigned addr=BX_SB16_IO; addr<=BX_SB16_IO+0xf; addr++) {
    BX_SB16_THIS devices->register_io_read_handler(this,
       read_handler,
       addr, "SB16");
    BX_SB16_THIS devices->register_io_write_handler(this,
       write_handler,
       addr, "SB16");
    }
  for (unsigned addr=BX_SB16_IOMPU; addr<=BX_SB16_IOMPU+3; addr++) {
    BX_SB16_THIS devices->register_io_read_handler(this,
       read_handler,
       addr, "SB16");
    BX_SB16_THIS devices->register_io_write_handler(this,
       write_handler,
       addr, "SB16");
    }

  writelog(BOTHLOG(3), "driver initialised, IRQ %d, IO %03x, DMA8 %d, DMA16 %d", 
	   BX_SB16_IRQ, BX_SB16_IO, BX_SB16_DMAL, BX_SB16_DMAH);

  // initialize the midi timer
  BX_SB16_THIS timer_handle = bx_pc_system.register_timer
    ( & BX_SB16_THISU , midi_timer, 500000 / 384, 1, 1);

  writelog(MIDILOG(6), "Timer initialized, handle %d", BX_SB16_THIS timer_handle);
  BX_SB16_THIS current_timer = 0;
}

  // the timer functions
void bx_sb16_c::midi_timer (void *this_ptr)
{
  ((bx_sb16_c *) this_ptr)->current_timer++;
  writelog(MIDILOG(6), "TIMER event - new timer: %d", ((bx_sb16_c *) this_ptr)->current_timer);
}

Bit32u bx_sb16_c::gettimer(void)
{
  return (BX_SB16_THIS current_timer);
}


  // the various IO handlers

  // The DSP/FM music part


// dsp_reset() resets the DSP after the sequence 1/0. Returns
// 0xaa on the data port
 
   void
bx_sb16_c::dsp_reset(Bit32u value)
{

  writelog(WAVELOG(4), "DSP Reset port write value %x", value);

  // just abort high speed mode if it is set
  if (DSP.dmahighspeed != 0)
    {
      DSP.dmahighspeed = 0;
      writelog(WAVELOG(4), "High speed mode aborted");
      return;
    }

  if ( (DSP.resetport == 1) &&
       (value == 0) )
    {

      // 1-0 sequences to reset port, do one of the following:
      // if in UART MIDI mode, abort it, don't reset
      // if in Highspeed mode (not SB16!), abort it, don't reset
      // otherwise reset

      if (DSP.midiuartmode != 0)
	{  // abort UART MIDI mode
	  DSP.midiuartmode = 0;
	  writelog(MIDILOG(4), "DSP UART MIDI mode aborted");
	  return;
	}

      // do the reset
      writelog(WAVELOG(4), "DSP resetting...");

      if (DSP.irqpending != 0)
	{
	  BX_SB16_THIS devices->pic->untrigger_irq(BX_SB16_IRQ);
	  writelog(WAVELOG(4), "DSP reset: IRQ untriggered");
	}
      if (DSP.dmamode != 0)
	{
	  //	  bx_pc_system.set_DRQ(BX_SB16_DMAL, 0);
	  writelog(WAVELOG(4), "DSP reset: DMA aborted");
	}

      DSP.resetport = 0;
      DSP.speaker = 0;
      DSP.irqpending = 0;
      DSP.midiuartmode = 0;
      DSP.prostereo = 0;

      DSP.dmamode = 0;
      DSP.dmafifo = 0;
      DSP.dmaoutput = 0;
      DSP.dmastereo = 0;
      DSP.dmasigned = 0;
      DSP.dmacount = 0;
      DSP.dmahighspeed = 0;

      DSP.currentsample = 0;
      DSP.currentsamplebytes = 0;

      DSP.dataout.reset();    // clear the buffers
      DSP.datain.reset();

      DSP.dataout.put(0xaa);  // acknowledge the reset
    }
  else
    DSP.resetport = value;
}


// dsp_dataread() reads the data port of the DSP

  Bit32u
bx_sb16_c::dsp_dataread()
{
  Bit8u value = 0xff;

  // if we are in MIDI UART mode, call the mpu401 part instead
  if (DSP.midiuartmode != 0)
    value = mpu_dataread();
  else
    {
      // default behaviour: if none available, return last byte again
      //  if (DSP.dataout.empty() == 0)
      DSP.dataout.get(&value);
    }

  writelog(WAVELOG(4), "DSP Data port read, result = %x", value);

  return(value);
}


// dsp_datawrite() writes a command or data to the data port

  void
bx_sb16_c::dsp_datawrite(Bit32u value)
{
  int bytesneeded;
  Bit8u mode;
  Bit16u length;

  writelog(WAVELOG(4), "DSP Data port write, value %x", value);

  // in high speed mode, any data passed to DSP is a sample
  if (DSP.dmahighspeed != 0)
    {
      dsp_getsamplebyte(value);
      return;
    }

  // route information to mpu401 part if in MIDI UART mode
  if (DSP.midiuartmode != 0)
    {
      mpu_datawrite(value);
      return;
    }

  if (DSP.datain.hascommand() == 1) // already a command pending, add to argument list
    {
      if (DSP.datain.put(value) == 0)
	{
	  writelog(WAVELOG(3), "DSP command buffer overflow for command %02x",
		   DSP.datain.currentcommand());
	}
    }
  else
    // no command pending, set one up
    {
      bytesneeded = 0;   // find out how many arguments the command takes
      switch (value)
	{   // fallbacks intended!
	case 0x10: 
	case 0x40: 
	case 0xe0: 
	case 0x38:
	  bytesneeded = 1; break;
	case 0x14: 
	case 0x16:
	case 0x17:
	case 0x41: 
	case 0x42: 
	case 0x48:
	case 0x74:
	case 0x75:
	case 0x76:
	case 0x77:
	case 0x80:
	case 0xe4: 
	  bytesneeded = 2; break;
	default:
	  switch (value >> 4)
	    {
	    case 0xb:
	    case 0xc:
	      bytesneeded = 3; break; // mode, lo(length) hi(length)
	    }
	}
      DSP.datain.newcommand(value, bytesneeded);
    }

  if (DSP.datain.commanddone() == 1) // command is complete, process it
    {
      writelog(WAVELOG(4), "DSP command %x with %d arg bytes",
	       DSP.datain.currentcommand(), DSP.datain.bytes());

      if ( ((DSP.datain.currentcommand() >> 4) == 0xb) ||
	   ((DSP.datain.currentcommand() >> 4) == 0xc) )
	{
	  DSP.datain.get(&mode);
	  DSP.datain.getw(&length);
	  dsp_dma(DSP.datain.currentcommand(), mode, length, 0);
	}
      else switch (DSP.datain.currentcommand())
	{		
	  // wave output commands - comments are the parameters for
	  // this command, and/or the output
	case 0x10: // direct mode
	  // 1: 8bit sample
	  DSP.datain.get(&mode);
	  DSP.currentsamplebytes = 1;
	  sendwavesample(mode);
	  DSP.currentsamplebytes = 0;
	  break;
	case 0x14: // uncomp'd, normal DMA
	  // 1,2: lo(length) hi(length)
	  DSP.datain.getw(&length);
	  dsp_dma(0xc0, 0x00, length, 0);
	  break;
	case 0x16: // 2-bit comp'd, normal DMA, no ref byte
	  // 1,2: lo(length) hi(length)
	  DSP.datain.getw(&length);
	  dsp_dma(0xc0, 0x00, length, 2);
	  break;
	case 0x17: // 2-bit comp'd, normal DMA, 1 ref byte
	  // 1,2: lo(length) hi(length)
	  DSP.datain.getw(&length);
	  dsp_dma(0xc0, 0x00, length, 2|8);
	  break;
	case 0x1c: // uncomp'd, auto DMA
	  // none
	  dsp_dma(0xc4, 0x00, DSP.blocklength, 0);
	  break;
	case 0x1f: // 2-bit comp'd, auto DMA, 1 ref byte
	  // none
	  dsp_dma(0xc4, 0x00, DSP.blocklength, 2|8);
	  break;
	case 0x74: // 4-bit comp'd, normal DMA, no ref byte
	  // 1,2: lo(length) hi(length)
	  DSP.datain.getw(&length);
	  dsp_dma(0xc0, 0x00, length, 4);
	  break;
	case 0x75: // 4-bit comp'd, normal DMA, 1 ref byte
	  // 1,2: lo(length) hi(length)
	  DSP.datain.getw(&length);
	  dsp_dma(0xc0, 0x00, length, 4|8);
	  break;
	case 0x76: // 3-bit comp'd, normal DMA, no ref byte
	  // 1,2: lo(length) hi(length)
	  DSP.datain.getw(&length);
	  dsp_dma(0xc0, 0x00, length, 3);
	  break;
	case 0x77: // 3-bit comp'd, normal DMA, 1 ref byte
	  // 1,2: lo(length) hi(length)
	  DSP.datain.getw(&length);
	  dsp_dma(0xc0, 0x00, length, 3|8);
	  break;
	case 0x7d: // 4-bit comp'd, auto DMA, 1 ref byte
	  // none
	  dsp_dma(0xc4, 0x00, DSP.blocklength, 4|8);
	  break;
	case 0x7f: // 3-bit comp'd, auto DMA, 1 ref byte
	  // none
	  dsp_dma(0xc4, 0x00, DSP.blocklength, 3|8);
	  break;

	case 0x90: // 8-bit auto DMA, highspeed
	  //none
	  dsp_dma(0xc4, 0x00, DSP.blocklength, 16);
	  break;

	case 0x91: // 8-bit normal DMA, highspeed
	  //none
	  dsp_dma(0xc0, 0x00, DSP.blocklength, 16);
	  break;
	  
	  // wave input commands, probably unnecessary to support them
	case 0x20: // ? direct mode
	  // o1: 8bit sample
	  DSP.dataout.put(0x80); // put a silence, for now.
	  break;
	case 0x24: // uncomp'd, normal DMA
	  // 1,2: lo(length) hi(length)
	  DSP.datain.getw(&length);
	  dsp_dma(0xc8, 0x00, length, 0);
	  break;
	case 0x2c: // uncomp'd, auto DMA
	  // none
	  dsp_dma(0xcc, 0x00, DSP.blocklength, 0);
	  break;
	  
	  // highspeed wave input
	case 0x98: // 8-bit auto DMA
	  //none
	  dsp_dma(0xcc, 0x00, DSP.blocklength, 16);
	  break;

	case 0x99: // 8-bit normal DMA
	  //none
	  dsp_dma(0xc8, 0x00, DSP.blocklength, 16);
	  break;
	  
	  // wave setup commands
	case 0x40: // set time constant
	  // 1: timeconstant
	  DSP.datain.get(&mode);
	  DSP.timeconstant = mode <<  8;
	  DSP.samplerate = (Bit32u) 256000000L / ((Bit32u) 65536L - (Bit32u) DSP.timeconstant);
	    break;
	case 0x41: // set t_c for input (fallback intended)
	case 0x42: // set t_c for output
	  // 1,2: hi(frq) lo(frq)
	  DSP.datain.getw1( &(DSP.samplerate) );
	  DSP.timeconstant = 65536 - (Bit32u) 256000000 / (Bit32u) DSP.samplerate ;
	  break; 
	case 0x48: // set block length
	  // 1,2: lo(blk len) hi(blk len)
	  DSP.datain.getw( &(DSP.blocklength) );
	  break;
	case 0x80: // silence period
	  // 1,2: lo(silence) hi(silence) (len in samples)
	  DSP.datain.getw(&length);
	  // only handled for VOC output so far
	  if (bx_options.sb16.wavemode == 2)
	    {
	      Bit8u temparray[3] = { length & 0xff, length >> 8, DSP.timeconstant >> 8 };
	      writevocblock(3, 0, temparray);
	    }
	  break;
	  
	case 0xa0: // switch to mono
	  // none
	  DSP.prostereo = 1;
	  break;
	case 0xa8: // switch to stereo
	  //// none
	  DSP.prostereo = 2;   // value only has meaning for SBPro transfers
	  break;

	case 0xd0: // pause 8 bit DMA transfer
	  // none
	  if (DSP.dmamode != 0)
	    bx_pc_system.set_DRQ(BX_SB16_DMAL, 0);
	  break;
	case 0xd1: // speaker on
	  // none
	  DSP.speaker = 1;
	  break;
	case 0xd3: // speaker off
	  // none
	  DSP.speaker = 0;
	  break;
	case 0xd4: // continue 8 bit DMA, see 0xd0
	  // none
	  if (DSP.dmamode != 0)
	    bx_pc_system.set_DRQ(BX_SB16_DMAL, 1);
	  break;
	case 0xd5: // pause 16 bit DMA
	  // none
	  if (DSP.dmamode != 0)
	    bx_pc_system.set_DRQ(BX_SB16_DMAH, 0);
	  break;
	case 0xd6: // continue 16 bit DMA, see 0xd5
	  // none
	  if (DSP.dmamode != 0)
	    bx_pc_system.set_DRQ(BX_SB16_DMAH, 0);
	  break;
	case 0xd8: // read speaker on/off (out ff=on, 00=off)
	  // none, o1: speaker; ff/00
	  DSP.dataout.put( (DSP.speaker == 1)?0xff:0x00);
	  break;
	case 0xd9: // stop 16 bit auto DMA
	  // none
	  if (DSP.dmamode != 0)
	    bx_pc_system.set_DRQ(BX_SB16_DMAH, 0);
	  DSP.dmamode = 0;
	  break;
	case 0xda: // stop 8 bit auto DMA
	  // none
	  if (DSP.dmamode != 0)
	    bx_pc_system.set_DRQ(BX_SB16_DMAL, 0);
	  DSP.dmamode = 0;
	  break;
	  
	  // Info commands
	case 0xe1: // get version, out 2 bytes (major, minor)
	  // none, o1/2: version major.minor
	  DSP.dataout.put(4);
	  if (DSP.dataout.put(11) == 0) 
	    {
	      writelog(WAVELOG(3), "DSP version couldn't be written - buffer overflow");
	    }
	  break;
	  
	  // MIDI commands
	case 0x30: // ? polling mode MIDI input
	  break;
	case 0x31: // ? interrupt mode MIDI input
	  break;
	case 0x34: // 0x34..0x37: UART mode MIDI output
	case 0x35: // UART mode MIDI input/output
	case 0x36: // UART polling mode MIDI IO with time stamp
	case 0x37: // UART interrupt mode MIDI IO with time stamp
	           // Fallbacks intended - all set the midi uart mode
	  DSP.midiuartmode = 1;
	  break;
	case 0x38: // MIDI output
	  DSP.datain.get(&mode);
	  // route to mpu401 part
	  mpu_datawrite(mode);
	  break;
	  
	  // undocumented commands. Just guessing!
	case 0xe0: // used and expected by the CL diagnose.exe
	  DSP.dataout.put(0x55);
	  break;
	case 0xe4: // used and expected by the CL diagnose.exe
	  DSP.dataout.put(0xaa);
	  break;
	case 0xf2: // Trigger IRQ on SB before Pro
	  DSP.dataout.put(0xaa);
	  DSP.irqpending = 1;
	  MIXER.reg[0x82] != 1; // reg 82 shows the kind of IRQ
	  BX_SB16_THIS devices->pic->trigger_irq(BX_SB16_IRQ);
	  break;
	  
	default: // unknown command
	  writelog(WAVELOG(3), "unknown DSP command %x, ignored",
		   DSP.datain.currentcommand());
	  break;
	}
      DSP.datain.clearcommand();
      DSP.datain.flush();
    }
}

// dsp_dma() initiates all kinds of dma transfers

  void
bx_sb16_c::dsp_dma(Bit8u command, Bit8u mode, Bit16u length, Bit8u comp)
{
  // command: 8bit, 16bit, in/out, single/auto, fifo 
  // mode: mono/stereo, signed/unsigned
  //   (for info on command and mode see sound blaster orogrammer's manual, 
  //    cmds bx and cz)
  // length: number of samples - not number of bytes
  // comp: bit-coded are: type of compression; ref-byte; highspeed
  //       D0..D2: 0=none, 2,3,4 bits ADPCM
  //       D3: ref-byte
  //       D6: highspeed

  writelog( WAVELOG(4), "DMA initialized. Cmd %02x, mode %02x, length %d, comp %d",
	    command, mode, length, comp);

  if ( (command >> 4) == 0xb )  // 0xb? = 16 bit DMA
    {
    DSP.dmabits = 16;
    DSP.dmabps = 2;
    }
  else                          // 0xc? = 8 bit DMA
    {
    DSP.dmabits = 8;
    DSP.dmabps = 1;
    }

  command &= 0x0f;
  DSP.dmaoutput = 1 - (command >> 3);    // 1=output, 0=input
  DSP.dmamode = 1 + (command >> 2) & 1;  // 0=none, 1=normal, 2=auto
  DSP.dmafifo = (command >> 1) & 1;      // ? not sure what this is

  DSP.dmastereo = (mode >> 5) & 1;

  if (DSP.dmastereo != 0)
    DSP.dmabps *= 2;

  DSP.dmasigned = (mode >> 4) & 1;
  DSP.dmacount = (length + 1) * DSP.dmabps - 1;
  DSP.dmahighspeed = (comp >> 4) & 1;

  DSP.currentsamplebytes = 0;

  if (bx_options.sb16.wavemode == 2)    // write the data to the voc file
    {
      Bit8u temparray[12] = 
        {DSP.samplerate & 0xff, DSP.samplerate >> 8, 0, 0,
	 DSP.dmabits, DSP.dmastereo + 1, 0, 0, 0, 0, 0, 0 };
      switch (comp & 7)
	{
	case 2: 
	  temparray[7] = 3;
	  break;
	case 3: 
	  temparray[7] = 2;
	  break;
	case 4:
	  temparray[7] = 1;
	  break;
	}
      writevocblock(9, -DSP.dmacount, temparray);
    }

  bx_pc_system.set_DRQ(BX_SB16_DMAL, 1);
}


// dsp_bufferstatus() checks if the DSP is ready for data/commands

  Bit32u
bx_sb16_c::dsp_bufferstatus()
{
  Bit32u result = 0x7f;

         // MSB set -> not ready for commands
  if (DSP.datain.full() == 1) result |= 0x80;

  writelog(WAVELOG(4), "DSP Buffer status read, result %x", result);

  return(result);
}


// dsp_status() checks if the DSP is ready to send data

  Bit32u
bx_sb16_c::dsp_status()
{
  Bit32u result = 0x7f;


  // read might be to acknowledge IRQ
  if ( DSP.irqpending != 0 )
    {
      DSP.irqpending = 0;
      MIXER.reg[0x82] &= (~0x01);
      writelog( WAVELOG(4), "8-bit DMA or SBMIDI IRQ acknowledged");
      if (MIXER.reg[0x82] == 0)
	BX_SB16_THIS devices->pic->untrigger_irq(BX_SB16_IRQ);
    }

  // if buffer is not empty, there is data to be read
  if (DSP.dataout.empty() == 0) result |= 0x80;

  writelog(WAVELOG(4), "DSP output status read, result %x", result);

  return(result);

}

// dsp_irq16ack() notifies that the 16bit DMA IRQ has been acknowledged
  Bit32u
bx_sb16_c::dsp_irq16ack()
{
  Bit32u result = 0xff;
  
  if ( DSP.irqpending != 0 )
    {
      DSP.irqpending = 0;
      MIXER.reg[0x82] &= (~0x02);
      if (MIXER.reg[0x82] == 0)
	BX_SB16_THIS devices->pic->untrigger_irq(BX_SB16_IRQ);
      writelog( WAVELOG(4), "16-bit DMA IRQ acknowledged");
    }
  else
    writelog( WAVELOG(3), "16-bit DMA IRQ acknowledged but not active!");

  return result;
}


  // the DMA handlers

      /* DMA input and outbut, 8 and 16 bit */
  void
bx_sb16_c::dma_write8(Bit8u *data_byte)
{
  writelog( WAVELOG(3), "8-bit DMA read not supported. Returning silence.");
  if (DSP.dmasigned == 0)
    *data_byte = 0x80;
  else
    *data_byte = 0;
}

  void
bx_sb16_c::dsp_getsamplebyte(Bit8u value)
{
  DSP.currentsamplebytes++;
  DSP.currentsample <<= 8;
  DSP.currentsample |= value;
  if (DSP.currentsamplebytes == DSP.dmabps)
    {
      if (bx_options.sb16.wavemode == 1)
	sendwavesample(DSP.currentsample);
      else if (bx_options.sb16.wavemode == 2)
	fwrite(&DSP.currentsample, 1, DSP.currentsamplebytes, WAVEDATA);
      DSP.currentsamplebytes = 0;
    DSP.currentsample = 0;
    }
}

  void
bx_sb16_c::dma_read8(Bit8u *data_byte)
{
  if (DSP.dmacount % 100 == 0) // otherwise it's just too many lines of log
    writelog( WAVELOG(5), "Received 8-bit DMA %2x, %d remaining ",
	      *data_byte, DSP.dmacount);
  DSP.dmacount--;

  dsp_getsamplebyte(*data_byte);

  if (DSP.dmacount == 0xffff) // last byte received
    {
      //unset DRQ line if it's single cycle DMA
      if (DSP.dmamode != 2)
	bx_pc_system.set_DRQ(BX_SB16_DMAL, 0);
      writelog( WAVELOG(4), "DMA transfer done, triggering IRQ");

      if (bx_options.sb16.wavemode == 2)
	fflush(WAVEDATA);
      else if (bx_options.sb16.wavemode == 1)
	sendwavepacket();

      //generate IRQ - note even for 16 bit sound, we are using
      //8 bit DMA, so trigger the 8bit IRQ
      //
      //      if (DSP.dmabits == 8)
        MIXER.reg[0x82] |= 1;
	//      else
	//	MIXER.reg[0x82] |= 2;
      BX_SB16_THIS devices->pic->trigger_irq(BX_SB16_IRQ);
      DSP.irqpending = 1;

      //if auto-DMA, reinitialize
      if (DSP.dmamode == 2)
	{
	  writelog( WAVELOG(4), "auto-DMA reinitializing");
	  DSP.dmacount = (DSP.blocklength + 1) * DSP.dmabps - 1;
	  bx_pc_system.set_DRQ(BX_SB16_DMAL, 1);
	}
      else
	DSP.dmamode = 0;     
    }
}

  void
bx_sb16_c::dma_write16(Bit16u *data_word)
{
  writelog( WAVELOG(3), "16-bit DMA read not supported. Returning silence.");
  if (DSP.dmasigned == 0)
    *data_word = 0x8000;
  else
    *data_word = 0;
}

  void
bx_sb16_c::dma_read16(Bit16u *data_word)
{
  writelog( WAVELOG(5), "Received 16-bit DMA %4x", *data_word);
  dsp_getsamplebyte(*data_word & 0xff);
  dsp_getsamplebyte(*data_word >> 8);
}


  // the mixer, supported type is CT1745 (as in an SB16)

  void
bx_sb16_c::mixer_writedata(Bit32u value)
{

  if (MIXER.regindex >= BX_SB16_MIX_REG)
    return;   // index out of range

  // store the value
  MIXER.reg[MIXER.regindex] = value;

  // do some action depending on what register was written
  switch (MIXER.regindex) 
    {
    case 0:  // initialize mixer      

      writelog(BOTHLOG(4), "Initializing mixer...");
      for (int i=0; i<BX_SB16_MIX_REG; i++)
	MIXER.reg[i] = 0;
      MIXER.reg[0x80] = 2;  // IRQ 5
      MIXER.reg[0x81] = 2;  // 8-bit DMA 1, no 16-bit DMA

      set_irq_dma();        // set the IRQ and DMA

      MIXER.regindex = 0;   // next mixer register read is register 0
      break;

    case 0x80: // IRQ mask
    case 0x81: // DMA mask
      set_irq_dma(); // both 0x80 and 0x81 handled
      break;

    // Note: some registers are bit-mapped to others. This should be
    // reflected here, in case somebody uses this to find out the
    // version of the DSP
    // These registers are 0x04, 0x0a, 0x22, 0x26, 0x28, 0x2e

    }
}

  Bit32u
bx_sb16_c::mixer_readdata()
{
  return(MIXER.reg[MIXER.regindex]);
  writelog(BOTHLOG(4), "Mixer register %02x set to %02x",
	   MIXER.regindex, MIXER.reg[MIXER.regindex]);
}

  void
bx_sb16_c::mixer_writeregister(Bit32u value)
{
  MIXER.regindex = value;
}


  void
bx_sb16_c::set_irq_dma()
{
  int newirq;

  // set the IRQ according to the value in mixer register 0x80
  switch (MIXER.reg[0x80]) 
    {
    case 1:
      newirq = 2;
      break;
    case 2:
      newirq = 5;
      break;
    case 4:
      newirq = 7;
      break;
    case 8:
      newirq = 10;
      break;
    default:
      newirq = 5;
      writelog(BOTHLOG(3), "Bad value %02x in mixer register 0x80. IRQ set to %d", 
	       MIXER.reg[0x80], newirq);
      MIXER.reg[0x80] = 2;
    }
  if (newirq != BX_SB16_IRQ)   // a different IRQ was set - shouldn't be!
    {
      /* unregister_irq not implemented yet
      if (BX_SB16_IRQ > 0)
        BX_SB16_THIS devices->unregister_irq(BX_SB16_IRQ);
	*/
      
      BX_SB16_IRQ = newirq;
      BX_SB16_THIS devices->register_irq(BX_SB16_IRQ,"SB16");
    }

  // set the 8 bit DMA
  switch (MIXER.reg[0x81] & 0x0f)
    {
    case 1: 
      BX_SB16_DMAL = 0;
      break;
    case 2:
      BX_SB16_DMAL = 1;
      break;
    case 8:
      BX_SB16_DMAL = 3;
      break;
    default:
      BX_SB16_DMAL = 1;
      writelog(BOTHLOG(3), "Bad value %02x in mixer register 0x81. DMA8 set to %d",
	       MIXER.reg[0x81], BX_SB16_DMAL);
      MIXER.reg[0x81] &= (~0x0f);
      MIXER.reg[0x81] |= 0x02;
    }

  // and the 16 bit DMA
  switch (MIXER.reg[0x81] >> 4)
    {
    case 0:
      BX_SB16_DMAH = 0;
      break;
    case 2:
      BX_SB16_DMAH = 5;
      break;
    case 4:
      BX_SB16_DMAH = 6;
      break;
    case 8:
      BX_SB16_DMAH = 7;
      break;
    default:
      BX_SB16_DMAH = 0;
      writelog(BOTHLOG(3), "Bad value %02x in mixer register 0x81. DMA16 set to %d",
	       MIXER.reg[0x81], BX_SB16_DMAH);
      MIXER.reg[0x81] &= (~0xf0);
      //      MIXER.reg[0x81] |= 2; // no default DMA16!
    }

  writelog(BOTHLOG(4), "Resources set to I%d D%d H%d", 
	   BX_SB16_IRQ, BX_SB16_DMAL, BX_SB16_DMAH);

}


  // now the MPU 401 part


// the MPU 401 status port shows if input or output are ready
// Note that the bits are inverse to their meaning

  Bit32u
bx_sb16_c::mpu_status()
{
  Bit32u result = 0;

  if (MPU.datain.full() == 1) result |= 0x40;	// no more output
  if (MPU.dataout.empty() == 1) result |= 0x80;	// no input available

  writelog(MIDILOG(4), "MPU status port, result %02x", result);
 
  return(result);
}


// the MPU 401 command port

  void
bx_sb16_c::mpu_command(Bit32u value)
{

  int bytesneeded;

  if (MPU.cmd.hascommand() == 1) // already a command pending, abort that one
    {
      if ( (MPU.cmd.currentcommand() != value) ||
	   (MPU.cmd.commanddone() == 0) )
	     // it's a different command, or the old one isn't done yet, abort it
	{
	  MPU.cmd.clearcommand();
	  MPU.cmd.flush();
	}
      // if it's the same one, and we just completed the argument list,
      // we leave it as it is and process it here
    }

  if (MPU.cmd.hascommand() == 0)  // no command pending, set one up
    {
      bytesneeded = 0;
      if ( (value >> 4) == 14) bytesneeded = 1;
      MPU.cmd.newcommand(value, bytesneeded);
    }

  if (MPU.cmd.commanddone() == 1) // command is complete, process it
    {
      switch (MPU.cmd.currentcommand())
	{
	case 0x3f:
	  writelog(MIDILOG(5), "MPU cmd: UART mode on");
	  MPU.uartmode=1;
	  MPU.irqpending=1;
	  MIXER.reg[0x82] |= 4;
	  BX_SB16_THIS devices->pic->trigger_irq(BX_SB16_IRQ);
	  break;

        case 0xff:
	  writelog(MIDILOG(4), "MPU cmd: Master reset of device");
	  MPU.uartmode=0;
	  for (int i=0; i<16; i++)
	    {
	      MPU.banklsb[i] = 0;
	      MPU.bankmsb[i] = 0;
	      MPU.program[i] = 0;
	    }
	  MPU.cmd.reset();
	  MPU.dataout.reset();
	  MPU.datain.reset();
	  MPU.midicmd.reset();
	  break;
	default:
	  writelog(MIDILOG(3), "MPU cmd: unknown command %02x ignored",
		   MPU.cmd.currentcommand());
	}

      // Need to put an MPU_ACK into the data port if command successful
      // we'll fake it even if we didn't process the command, so as to 
      // allow detection of the MPU 401.
      if (MPU.dataout.put(0xfe) == 0)
	writelog(MIDILOG(3), "MPU_ACK error - output buffer full");
      MPU.cmd.clearcommand(); // clear the command from the buffer
    }
}


// MPU 401 data port/read: contains an MPU_ACK after receiving a command
// Will contain other data as well when other than UART mode is supported

  Bit32u
bx_sb16_c::mpu_dataread()
{
    Bit8u res8bit;
    Bit32u result;

    // also acknowledge IRQ?
    if ( MPU.irqpending != 0 )
      {
	MPU.irqpending = 0;
	MIXER.reg[0x82] &= (~4);
	if (MIXER.reg[0x82] == 0)
	  BX_SB16_THIS devices->pic->untrigger_irq(BX_SB16_IRQ);
      }

    if ( MPU.dataout.get(&res8bit) == 0)
      {
	writelog(MIDILOG(3), "MPU data port not ready - no data in buffer");
	result = 0xff;
      }
    else
      result = (Bit32u) res8bit;

    writelog(MIDILOG(4), "MPU data port, result %02x", result);

    return(result);
}


// MPU 401 data port/write: This is where the midi stream comes from,
// as well as arguments to any pending command

  void
bx_sb16_c::mpu_datawrite(Bit32u value)
{
  writelog(MIDILOG(4), "write to MPU data port, value %02x", value);

  if (MPU.cmd.hascommand() == 1)
    { // there is a command pending, add arguments to it
      if (MPU.cmd.put(value) == 0)
	writelog(MIDILOG(3), "MPU Command arguments too long - buffer full");
      if (MPU.cmd.commanddone() == 1)
	BX_SB16_THIS mpu_command(MPU.cmd.currentcommand());
    }
  else if (MPU.uartmode == 0)
    {
      // Hm? No UART mode, but still data? Maybe should send it
      // to the command port...  Only SBMPU401.EXE does this...
      writelog(MIDILOG(4), "MPU Data %02x received but no UART mode. Assuming it's a command.", value);
      mpu_command(value);
      return;
    }
  else // no MPU command pending, in UART mode, this has to be midi data
    {
      // first, find out if it is a midi command or midi data
      Boolean ismidicommand = 0;
      if (value >= 0x80)
	{  // bit 8 usually denotes a midi command, except for CSYSEX
	  ismidicommand = 1;
	  if ( (value == 0xf7) && (MPU.midicmd.currentcommand() == 0xf0) )
	    // ...except if it is a continuing SYSEX message, then it just
	    // denotes the end of a SYSEX chunk, not the start of a message
	    {
	      ismidicommand = 0;     // first, it's not a command
	      MPU.midicmd.newcommand(MPU.midicmd.currentcommand(), 
				    MPU.midicmd.bytes());
	      // Then, set needed bytes to current buffer
	      // because we didn't know the length before
	    }
	}
      
      if (ismidicommand == 1)
	{  // this is a command, check if an old one is pending
	  if (MPU.midicmd.hascommand() == 1)
	    {
	      writelog(MIDILOG(3), "Midi command %02x incomplete, has %d of %d bytes.", 
		       MPU.midicmd.currentcommand(), MPU.midicmd.bytes(), 
		       MPU.midicmd.commandbytes() );
	      // write as much as we can. Should we do this?
	      processmidicommand(0);
	      // clear the pending command
	      MPU.midicmd.clearcommand();
	      MPU.midicmd.flush();
	    }
	  
	  // find the number of arguments to the command
	  static signed eventlength[] = { 2, 2, 2, 2, 1, 1, 2, 255};
      	      // note - length 255 commands have unknown length
	  MPU.midicmd.newcommand(value, eventlength[ ( (value & 0xf0) >> 4 ) - 8 ]);
	}
      else  // no command, just arguments to the old command
	{
	  if (MPU.midicmd.hascommand() == 0)
	    {  // no command pending, ignore the data
	      writelog(MIDILOG(3), "Midi data %02x received, but no command pending?", value);
	      return;
	    }

	  // just some data to the command
	  if (MPU.midicmd.put(value) == 0)
	    writelog(MIDILOG(3), "Midi buffer overflow!");
	  if (MPU.midicmd.commanddone() == 1)
	    {
	      // the command is complete, process it
	      writelog(MIDILOG(5), "Midi command %02x complete, has %d bytes.", 
		       MPU.midicmd.currentcommand(), MPU.midicmd.bytes() );
	      processmidicommand(0);
	      // and remove the command from the buffer
	      MPU.midicmd.clearcommand();
	      MPU.midicmd.flush();
	    }
	}
    }
  return;
}


// The emulator port/read: See if commands were successful

  Bit32u
bx_sb16_c::emul_read()
{
    Bit8u res8bit;
    Bit32u result;

    if ( EMUL.datain.get(&res8bit) == 0)
      {
	writelog(3, "emulator port not ready - no data in buffer");
	result = 0x00;
      }
    else
      result = (Bit32u) res8bit;

    writelog(4, "emulator port, result %02x", result);

    return(result);
}

// Emulator port/write: Changing instrument mapping etc.

  void
bx_sb16_c::emul_write(Bit32u value)
{

  Bit8u value8;

  writelog(4, "write to emulator port, value %02x", value);

  if (EMUL.dataout.hascommand() == 0)   // no command pending, set it up
    {
      static signed char cmdlength[] = { 0, 0, 4, 2, 6, 1, 0 };
      writelog(4, "emulator command %02x, needs %d arguments",
	       value, cmdlength[value]);
      EMUL.dataout.newcommand(value, cmdlength[value]);
      EMUL.datain.put(0xfe);
    }
  else
    EMUL.dataout.put(value);    // otherwise just add data

  if (EMUL.dataout.commanddone() == 1)
    { // process the command
      writelog(4, "executing emulator command %02x with %d arguments",
	       EMUL.dataout.currentcommand(), EMUL.dataout.bytes());
      switch (EMUL.dataout.currentcommand())
	{
	case 0: // reinit of emulator
	  writelog(4, "Emulator reinitialized");
	  
	  EMUL.remaps = 0;

	  EMUL.dataout.reset();
	  EMUL.datain.reset();
	  EMUL.datain.put(0xfe);
	  break;
	case 1: // dummy command to reset state of emulator port
	  break;
	case 2: // map bank
	  EMUL.dataout.get (& (EMUL.remaplist[EMUL.remaps].oldbankmsb));
	  EMUL.dataout.get (& (EMUL.remaplist[EMUL.remaps].oldbanklsb));
	  EMUL.remaplist[EMUL.remaps].oldprogch = 0xff;
	  EMUL.dataout.get (& (EMUL.remaplist[EMUL.remaps].newbankmsb));
	  EMUL.dataout.get (& (EMUL.remaplist[EMUL.remaps].newbanklsb));
	  EMUL.remaplist[EMUL.remaps].newprogch = 0xff;
	  EMUL.datain.put(4);
	  writelog(4, "Map bank command received, from %d %d to %d %d",
		   EMUL.remaplist[EMUL.remaps].oldbankmsb,
		   EMUL.remaplist[EMUL.remaps].oldbanklsb,
		   EMUL.remaplist[EMUL.remaps].newbankmsb,
		   EMUL.remaplist[EMUL.remaps].newbanklsb);
	  EMUL.remaps++;
	  break;
	case 3: // map program change
	  EMUL.remaplist[EMUL.remaps].oldbankmsb = 0xff;
	  EMUL.remaplist[EMUL.remaps].oldbanklsb = 0xff;
	  EMUL.dataout.get (& (EMUL.remaplist[EMUL.remaps].oldprogch));
	  EMUL.remaplist[EMUL.remaps].newbankmsb = 0xff;
	  EMUL.remaplist[EMUL.remaps].newbanklsb = 0xff;
	  EMUL.dataout.get (& (EMUL.remaplist[EMUL.remaps].newprogch));
	  EMUL.datain.put(2);
	  writelog(4, "Map program change received, from %d to %d",
		   EMUL.remaplist[EMUL.remaps].oldprogch,
		   EMUL.remaplist[EMUL.remaps].newprogch);
	  EMUL.remaps++;
	  break;
	case 4: // map bank and program change
	  EMUL.dataout.get (& (EMUL.remaplist[EMUL.remaps].oldbankmsb));
	  EMUL.dataout.get (& (EMUL.remaplist[EMUL.remaps].oldbanklsb));
	  EMUL.dataout.get (& (EMUL.remaplist[EMUL.remaps].oldprogch));
	  EMUL.dataout.get (& (EMUL.remaplist[EMUL.remaps].newbankmsb));
	  EMUL.dataout.get (& (EMUL.remaplist[EMUL.remaps].newbanklsb));
	  EMUL.dataout.get (& (EMUL.remaplist[EMUL.remaps].newprogch));
	  EMUL.datain.put(6);
	  writelog(4, "Complete remap received, from %d %d %d to %d %d %d",
		   EMUL.remaplist[EMUL.remaps].oldbankmsb,
		   EMUL.remaplist[EMUL.remaps].oldbanklsb,
		   EMUL.remaplist[EMUL.remaps].oldprogch,
		   EMUL.remaplist[EMUL.remaps].newbankmsb,
		   EMUL.remaplist[EMUL.remaps].newbanklsb,
		   EMUL.remaplist[EMUL.remaps].newprogch);

	  EMUL.remaps++;
	  break;
	case 5: EMUL.dataout.get(&value8); // dump emulator state
	  switch (value8)
	    {
	    case 0: EMUL.datain.puts("SB16 Emulator for Bochs\n");
	      break;
	    case 1: EMUL.datain.puts("UART mode=%d\n", MPU.uartmode);
	      break;
	    case 2: EMUL.datain.puts("timer=%d\n", BX_SB16_THIS current_timer);
	      break;
	    case 3: EMUL.datain.puts("%d remappings active\n", EMUL.remaps);
	      break;
	    case 4: EMUL.datain.puts("Hardware config A%3x I%d D%d H%d T%d P%3x\n",
				BX_SB16_IO, BX_SB16_IRQ, BX_SB16_DMAL,
				BX_SB16_DMAH, 6, BX_SB16_IOMPU);
	    default: EMUL.datain.puts("no info. Only 0..4 have values.\n");
	    }
	  break;
	case 6: // close midi and wave files and/or output
	  if (bx_options.sb16.midimode == 2)
	    {
	      finishmidifile();
	      fclose(MIDIDATA);
	    }
	  else if (bx_options.sb16.midimode == 1)
	    closemidioutput();
	  bx_options.sb16.midimode = 0;
	  
	  if (bx_options.sb16.wavemode == 2)
	    {
	      finishvocfile();
	      fclose(WAVEDATA);
	    }
	  else
	    closewaveoutput();
	  bx_options.sb16.wavemode = 0;

	  break;
	case 7: // clear bank/program mappings
	  EMUL.remaps = 0;
	  writelog(4, "Bank/program mappings cleared.");
	  break;
	}
      EMUL.dataout.clearcommand();
      EMUL.dataout.flush();
    }
}


/* Handlers for the midi commands/midi file output */

// Write the header of the midi file. Track length is 0x7fffffff 
// until we know how long it's really going to be

  void
bx_sb16_c::initmidifile()
{
  struct {
    Bit8u chunk[4];
    Bit32u chunklen;  // all in BIG Endian!
    Bit16u smftype; 
    Bit16u tracknum;
    Bit16u timecode;  // 0x8000 + deltatimesperquarter
    } midiheader = 
#ifdef BX_LITTLE_ENDIAN
      { "MThd", 0x06000000, 0x0100, 0x0100, 0x8002 };
#else
      { "MThd", 6, 1, 1, 0x280 };
#endif

  struct {
    Bit8u chunk[4];
    Bit32u chunklen;
    Bit8u data[15];
    } trackheader = 
#ifdef BX_LITTLE_ENDIAN
      { "MTrk", 0xffffff7f, 
#else
      { "MTrk", 0x7fffffff,
#endif
	{ 0x00,0x2f,0x51,3,0x07,0xa1,0x20,    // set tempo 120 (0x7a120 us per quarter)
	  0x00,0x2f,0x58,4,4,2,0x18,0x08 }};  // time sig 4/4
    
  fwrite(&midiheader, 1, sizeof midiheader, MIDIDATA );
  fwrite(&trackheader, 1, sizeof trackheader, MIDIDATA );
  return;
}

// write the midi command to the midi file
  
  void
bx_sb16_c::writemidicommand(int deltatime, int command, int length, Bit8u data[])
{
  int i;

  writedeltatime(deltatime);
  i = 0;
  fputc(command, MIDIDATA);
  if ( (command == 0xf0) || 
       (command == 0xf7) )    // write event length for sysex/meta events
    writedeltatime(length);

  fwrite(data, 1, length, MIDIDATA);
}


// process the midi command stored in MPU.midicmd.to the midi driver

  void
bx_sb16_c::processmidicommand(Boolean force)
{
  int i, channel;
  Bit8u value;
  Bit32u deltatime;
  Boolean needremap = 0;
  static Bit32u oldtime = 0xffffffff;

  channel = MPU.midicmd.currentcommand() & 0xf;

  /* We need to determine the time elapsed since the last command */

  if (oldtime == 0xffffffff) oldtime = BX_SB16_THIS gettimer();
 
  deltatime = BX_SB16_THIS gettimer() - oldtime;
  oldtime = BX_SB16_THIS gettimer();

  // we need to log bank changes and program changes
  if ( (MPU.midicmd.currentcommand() >> 4) == 0xc)
    {    // a program change
      value = MPU.midicmd.peek(0);
      writelog(MIDILOG(1), "* ProgramChange channel %d to %d",
	       channel, value);
      MPU.program[channel] = value;
      needremap = 1;
    }
  else if ( (MPU.midicmd.currentcommand() >> 4) == 0xb)
    {    // a control change, could be a bank change
      if (MPU.midicmd.peek(0) == 0)
	{  // bank select MSB
	  value = MPU.midicmd.peek(1);
	  writelog(MIDILOG(1), "* BankSelectMSB (%x %x %x) channel %d to %d",
		   MPU.midicmd.peek(0), MPU.midicmd.peek(1), MPU.midicmd.peek(2),
		   channel, value);
	  MPU.bankmsb[channel] = value;
	  needremap = 1;
	}
      else if (MPU.midicmd.peek(0) == 32)
	{  // bank select LSB
	  value = MPU.midicmd.peek(1);
	  writelog(MIDILOG(1), "* BankSelectLSB channel %d to %d",
		   channel, value);
	  MPU.banklsb[channel] = value;
	  needremap = 1;
	}
    }

  Bit8u temparray[256];
  i = 0;
  while (MPU.midicmd.empty() == 0)
    MPU.midicmd.get( &(temparray[i++]) );

  MIDIOUTPUT(deltatime, MPU.midicmd.currentcommand(), i, temparray);

  if ( (force == 0) && (needremap == 1) )
    // have to check the remap lists, and remap program change if necessary
    midiremapprogram(channel);
}


// check if a program change has to be remapped, and do it if necessary

   void 
bx_sb16_c::midiremapprogram(int channel)
{
  int bankmsb,banklsb,program;
  Bit8u commandbytes[2];

  bankmsb = MPU.bankmsb[channel];
  banklsb = MPU.banklsb[channel];
  program = MPU.program[channel];

  for (int i = 0; i < EMUL.remaps; i++)
    {
      if ( ( (EMUL.remaplist[i].oldbankmsb == bankmsb) ||
	     (EMUL.remaplist[i].oldbankmsb == 0xff) ) &&
	   ( (EMUL.remaplist[i].oldbanklsb == banklsb) ||
	     (EMUL.remaplist[i].oldbanklsb == 0xff) ) &&
	   ( (EMUL.remaplist[i].oldprogch == program) ||
	     (EMUL.remaplist[i].oldprogch == 0xff) ) )
	{
	  writelog(5, "Remapping instrument for channel %d", channel);
	  if ( (EMUL.remaplist[i].newbankmsb != bankmsb) &&
	       (EMUL.remaplist[i].newbankmsb != 0xff) )
	    {  // write control change bank msb
	      MPU.bankmsb[channel] = EMUL.remaplist[i].newbankmsb;
	      commandbytes[0] = 0;
	      commandbytes[1] = EMUL.remaplist[i].newbankmsb;
	      MIDIOUTPUT(0, 0xb0 | channel, 2, commandbytes);
	    }
	  if ( (EMUL.remaplist[i].newbanklsb != banklsb) &&
	       (EMUL.remaplist[i].newbanklsb != 0xff) )
	    {  // write control change bank lsb
	      MPU.banklsb[channel] = EMUL.remaplist[i].newbanklsb;
	      commandbytes[0] = 32;
	      commandbytes[1] = EMUL.remaplist[i].newbanklsb;
	      MIDIOUTPUT(0, 0xb0 | channel, 2, commandbytes);
	    }
	  if ( (EMUL.remaplist[i].newprogch != program) &&
	       (EMUL.remaplist[i].newprogch != 0xff) )
	    {  // write program change
	      MPU.program[channel] = EMUL.remaplist[i].newprogch;
	      commandbytes[0] = EMUL.remaplist[i].newprogch;
	      MIDIOUTPUT(0, 0xc0 | channel, 1, commandbytes);
	    }
	}
    }
}


// convert a number into a delta time coded value
   int 
bx_sb16_c::converttodeltatime(Bit32u deltatime, Bit8u value[4])
{
int i, count;
Bit8u outbytes[4];

count = 0;

if (deltatime <= 0)
  {
    count = 1;
    value[0] = 0;
  }
else 
  {
    while ( (deltatime > 0) && (count < 4) )   // split into parts
      {                                        // of seven bits
	outbytes[count++] = deltatime & 0x7f;
	deltatime >>= 7;
      };
    for (i=0; i<count; i++)                      // reverse order and
      value[i] = outbytes[count - i - 1] | 0x80; // set eighth bit on
    value[count - 1] &= 0x7f;                    // all but last byte
  }
return count;
}


// write a delta time coded value to the midi file
   void
bx_sb16_c::writedeltatime(Bit32u deltatime)
{
Bit8u outbytes[4];
int count,i;

count = converttodeltatime(deltatime, outbytes);

for (i=0;i<count;i++)
  fputc(outbytes[i], MIDIDATA );
}


// close the midi file, and set the track length accordingly

  void
bx_sb16_c::finishmidifile()
{
  fpos_t tracklen;
  struct { 
    Bit8u delta, statusbyte, metaevent, length;
  } metatrackend = { 0, 0xff, 0x2f, 0 };

     // Meta event track end (0xff 0x2f 0x00) plus leading delta time
  fwrite(&metatrackend, 1, sizeof metatrackend, MIDIDATA );

  fgetpos(MIDIDATA, &tracklen);
  tracklen -= 22;    // subtract the midi file and track header

  fseek(MIDIDATA, 22 - 4, SEEK_SET);

     // value has to be in big endian
#ifdef BX_LITTLE_ENDIAN
  tracklen = (tracklen << 24) | (tracklen >> 24) |
        ((tracklen & 0x00ff0000) >> 8) |
        ((tracklen & 0x0000ff00) << 8);
#endif
  fwrite(&tracklen, 4, 1, MIDIDATA);

  return;
}



/* Handlers for the voc file output */

// Write the header of the voc file.

  void
bx_sb16_c::initvocfile()
{
  struct { 
    char id[20];
    Bit16u headerlen;  // All in LITTLE Endian!
    Bit16u version;
    Bit16u magic;
  } vocheader =
    { "Creative Voice File\032",
#ifdef BX_LITTLE_ENDIAN
      0x1a, 0x0114, 0x111f };
#else
      0x1a00, 0x1401, 0x1f11 };
#endif
    
  fwrite(&vocheader, 1, sizeof vocheader, WAVEDATA);
}

// write one block to the voc file
// abs(length) is the number of additional bytes after the header.
// if length<0, the data is written by another function and not
// yet in the data[] array.
  void
bx_sb16_c::writevocblock(int block, Bit32s length, Bit8u data[])
{
  int i;
  Bit8u lengthbytes[3];
  Bit32u headerlen[10] = { 0, 2, 0, 3, 2, 0, 2, 0, 4, 12 };

  if (block > 9)
    {
      writelog( WAVELOG(3), "VOC Block %d not recognized, ignored.", block);
      return;
    }

  fputc(block, WAVEDATA);

  i = headerlen[block] + abs(length);
#ifdef BX_LITTLE_ENDIAN
  fwrite(&i, 1, 3, WAVEDATA);       // write the length in 24-bit little endian
#else
  lengthbytes[0] = i & 0xff; i >>= 8;
  lengthbytes[1] = i & 0xff; i >>= 8;
  lengthbytes[2] = i & 0xff;
#endif
  if (length < 0)
    length = 0;    // only write the header in this case
  if (headerlen[block] + length > 0)
    fwrite(data, 1, headerlen[block] + length, WAVEDATA);
}


// close the voc file
  void
bx_sb16_c::finishvocfile()
{
  fputc(0, WAVEDATA);  // blocktype 0: end block
}


  // static IO port read callback handler
  // redirects to non-static class handler to avoid virtual functions

  Bit32u
bx_sb16_c::read_handler(void *this_ptr, Bit32u address, unsigned io_len)
{
#if !BX_USE_SB16_SMF
  bx_sb16_c *class_ptr = (bx_sb16_c *) this_ptr;

  return( class_ptr->read(address, io_len) );
}


  Bit32u
bx_sb16_c::read(Bit32u address, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif  // !BX_USE_SB16_SMF

  bx_sb16_read_t dsp_ports[BX_SB16_IOLEN] =
      {NULL,              // 2x0: FM Music Status Port
       NULL,              // 2x1: reserved (w: FM Music Data Port)
       NULL,              // 2x2: Advanced Music Status Port
       NULL,              // 2x3: reserved (w: Adv. FM Music Data Port)		break;
       NULL,              // 2x4: reserved (w: Mixer Register Port)
       mixer_readdata,    // 2x5: Mixer Data Port
       NULL,              // 2x6: reserved (w: DSP Reset)
       NULL,              // 2x7: reserved
       NULL,              // 2x8: FM Music Status Port (OPL-2)
       NULL,              // 2x9: reserved (w: FM Music Data Port)
       dsp_dataread,      // 2xa: DSP Read Data Port
       NULL,              // 2xb: reserved
       dsp_bufferstatus,  // 2xc: DSP Buffer Status Port
       NULL,              // 2xd: reserved
       dsp_status,        // 2xe: DSP Data Status Port
       dsp_irq16ack};     // 2xf: DSP Acknowledge 16bit DMA IRQ

  bx_sb16_read_t mpu_ports[BX_SB16_IOMPULEN] =
      {mpu_dataread,      // 3x0: MPU Data Port Read
       mpu_status,        // 3x1: MPU Status Port
       NULL,              // 3x2: reserved
       emul_read};        // 3x3: *Emulator* Port

// we support only byte access to the port
  if (io_len != 1)
    {
      writelog(3, "Read access to %03x not byte access, len=%d!",
	       address, io_len);
      return(0xff);
    }

  if ( (address >= BX_SB16_IO) && 
       (address < BX_SB16_IO + BX_SB16_IOLEN) )
    // DSP or FM ports
    {
      if ( dsp_ports[address - BX_SB16_IO] != NULL )
	return(dsp_ports[address - BX_SB16_IO]());
    }
  else if ( (address >= BX_SB16_IOMPU) &&
	    (address < BX_SB16_IOMPU + BX_SB16_IOMPULEN) )
    // MPU401 ports
    {
      if ( mpu_ports[address - BX_SB16_IOMPU] != NULL )
	return(mpu_ports[address - BX_SB16_IOMPU]());
    }

  // If we get here, the port wasn't valid
  writelog(3, "Read access to %03x for %d: unsupported port!", address, io_len);

  return(0xff);
}

  // static IO port write callback handler
  // redirects to non-static class handler to avoid virtual functions

  void
bx_sb16_c::write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len)
{
#if !BX_USE_SB16_SMF
  bx_sb16_c *class_ptr = (bx_sb16_c *) this_ptr;

  class_ptr->write(address, value, io_len);
}

  void
bx_sb16_c::write(Bit32u address, Bit32u value, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif  // !BX_USE_SB16_SMF

  bx_sb16_write_t dsp_ports[BX_SB16_IOLEN] =
      {NULL,              // 2x0: FM Music Register Port
       NULL,              // 2x1: FM Music Data Port
       NULL,              // 2x2: Advanced FM Music Register Port
       NULL,              // 2x3: Advanced FM Music Data Port
       mixer_writeregister,//2x4: Mixer Register Port
       mixer_writedata,   // 2x5: Mixer Data Port
       dsp_reset,         // 2x6: DSP Reset
       NULL,              // 2x7: reserved
       NULL,              // 2x8: FM Music Register Port (OPL-2)
       NULL,              // 2x9: FM Music Data Port
       NULL,              // 2xa: reserved (r: DSP Data Port)
       NULL,              // 2xb: reserved
       dsp_datawrite,     // 2xc: DSP Write Command/Data
       NULL,              // 2xd: reserved
       NULL,              // 2xe: reserved (r: DSP Buffer Status)
       NULL};             // 2xf: reserved

  bx_sb16_write_t mpu_ports[BX_SB16_IOMPULEN] =
      {mpu_datawrite,     // 3x0: MPU Command Port
       mpu_command,       // 3x1: MPU Data Port
       NULL,              // 3x2: reserved
       emul_write};       // 3x3: *Emulator* Port

  // we only support byte access to the prot
  if (io_len != 1)
	{
	  writelog(3, "Write access to %03x for %d to %02x: "
		      "not byte access!", address, io_len, value);
	  return;
	}

  if ( (address >= BX_SB16_IO) && 
       (address < BX_SB16_IO + BX_SB16_IOLEN) )
    {
      if ( dsp_ports[address - BX_SB16_IO] != NULL )
	{
	  dsp_ports[address - BX_SB16_IO](value);
	  return;
	}
    }
  else if ( (address >= BX_SB16_IOMPU) &&
	    (address < BX_SB16_IOMPU + BX_SB16_IOMPULEN) )
    {
      if ( mpu_ports[address - BX_SB16_IOMPU] != NULL )
	{
	  mpu_ports[address - BX_SB16_IOMPU](value);
	  return;
	}
    }

  // if we arrive here, the port is unsupported
  writelog(3, "Write access to %03x for %d to %02x: unsupported port!", 
	   address, io_len, value);

  return;
}

void bx_sb16_c::writelog(int loglevel, const char *str, ...)
{
// append a line to the log file, if desired
  if (bx_options.sb16.loglevel >= loglevel)
    {
        time_t timep = time(NULL);
        tm *t = localtime(&timep); 
	fprintf(LOGFILE, "SB16 %02d:%02d:%02d (%i): ", 
		t->tm_hour, t->tm_min, t->tm_sec, loglevel);
	va_list ap;
	va_start(ap, str);
	vfprintf(LOGFILE, str, ap);
	va_end(ap);
	fprintf(LOGFILE, "\n");
    }
}

// the round-robin FIFO buffers of the SB16
bx_sb16_buffer::bx_sb16_buffer()
{
  length = 0;		// total bytes in buffer
  head = 0;		// pointer to next slot available for new data
  tail = 0;		// pointer to next slot to be read from
  buffer = NULL;	// pointer to the actual data
}

void bx_sb16_buffer::init(int bufferlen)
{
  if (buffer != NULL)	// Was it initialized before?
    delete buffer;

  length = bufferlen;
  buffer = new Bit8u[length];
  if (buffer == NULL)
    length = 0;		// This will be checked later

  reset();
}

void bx_sb16_buffer::reset()
{
  head = 0;		// Reset the pointers
  tail = 0;

  clearcommand();       // no current command set
}

bx_sb16_buffer::~bx_sb16_buffer(void)
{
  if (buffer != NULL)
    delete buffer;

  buffer = NULL;
  length = 0;
}


// Report how many bytes are available
int bx_sb16_buffer::bytes(void)
{
  if (empty() != 0)
    return 0;   // empty / not initialized

  int bytes = head - tail;
  if (bytes < 0) bytes += length;
  return (bytes);
}

// This puts one byte into the buffer
Boolean bx_sb16_buffer::put(Bit8u data)
{
  if (full() != 0)
    return 0;	// buffer full

  buffer[head++] = data;	// Write data, and increase write pointer
  head %= length;		// wrap it around so it stays inside the data

  return 1;	// put was successful
}

// This writes a formatted string to the buffer
Boolean bx_sb16_buffer::puts(char *data, ...)
{

  if (data == NULL)
    return 0;  // invalid string

  char string[length];
  int index = 0;

  va_list ap;
  va_start(ap, data);
  vsnprintf(string, length, data, ap);
  va_end(ap);

  while (string[index] != 0)
    {
      if (put( (Bit8u) string[index]) == 0)
	return 0;  // buffer full
      index++;
    }
  return 1;
}

// This returns if the buffer is full, i.e. if a put will fail
Boolean bx_sb16_buffer::full(void)
{
  if (length == 0)
    return 1;   // not initialized

  if ( ((head + 1) % length) == tail)
    return 1;   // buffer full

  return 0;	// buffer has some space left
}

// This reads the next available byte from the buffer
Boolean bx_sb16_buffer::get(Bit8u *data)
{
  if (empty() != 0)
    { 
      // Buffer is empty. Still, if it was initialized, return
      // the last byte again.
      if ( length > 0 )
	(*data) = buffer[ (tail - 1) % length ];
      return 0;	// buffer empty
    }

  (*data) = buffer[tail++];	// read data and increase read pointer
  tail %= length;		// and wrap it around to stay inside the data

  return 1;	// get was successful
}

// Read a word in lo/hi order
Boolean bx_sb16_buffer::getw(Bit16u *data)
{
  Bit8u dummy;
  if (bytes() < 2)
    {
      if (bytes() == 1)
	{
	  get(&dummy);
	  *data = (Bit16u) dummy;
	}
      else
	dummy = 0;
      return 0;
    }
  get(&dummy);
  *data = (Bit16u) dummy;
  get(&dummy);
  *data |= ( (Bit16u) dummy ) << 8;
}

// Read a word in hi/lo order
Boolean bx_sb16_buffer::getw1(Bit16u *data)
{
  Bit8u dummy;
  if (bytes() < 2)
    {
      if (bytes() == 1)
	{
	  get(&dummy);
	  *data = ( (Bit16u) dummy ) << 8;
	}
      else
	dummy = 0;
      return 0;
    }
  get(&dummy);
  *data = ( (Bit16u) dummy ) << 8;
  get(&dummy);
  *data |= (Bit16u) dummy;
}

// This returns if the buffer is empty, i.e. if a get will fail
Boolean bx_sb16_buffer::empty(void)
{
  if (length == 0)
    return 1;   // not inialized

  if (head == tail)
    return 1;   // buffer empty

  return 0;	// buffer contains data
}


// Flushes the buffer
void bx_sb16_buffer::flush(void)
{
  tail = head;
  return;
}

// Peeks ahead in the buffer
// Warning: No checking if result is valid. Must call bytes() to check that!
Bit8u bx_sb16_buffer::peek(int offset)
{
  return buffer[ (tail + offset) % length ];
}

// Set a new active command
void bx_sb16_buffer::newcommand(Bit8u newcmd, int bytes)
{
  command = newcmd;
  havecommand = 1;
  bytesneeded = bytes;
}

// Return the currently active command
Bit8u bx_sb16_buffer::currentcommand(void)
{
  return command;
}

// Clear the active command
void bx_sb16_buffer::clearcommand(void)
{
  command = 0;
  havecommand = 0;
  bytesneeded = 0;
}

// return if the command has received all necessary bytes
Boolean bx_sb16_buffer::commanddone(void)
{
  if (hascommand() == 0)
    return 0;  // no command pending - not done then

  if (bytes() >= bytesneeded)
    return 1;  // yes, it's done

  return 0;    // no, it's not
}

// return if there is a command pending
Boolean bx_sb16_buffer::hascommand(void)
{
  return havecommand;
}

int bx_sb16_buffer::commandbytes(void)
{
  return bytesneeded;
}
