// Copyright Notice
// ================
// BOCHS is Copyright 1994-1998 by Kevin P. Lawton.
//
// This file (SOUNDLNX.CC) Copyright 1998-1999 by Josef Drexler
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
// 
//     EMail:    bochs@world.std.com


// This file implements the sound output to the Linux devices /dev/midi??
// for music and /dev/dsp for audio output.


#include "bochs.h"


#if BX_USE_SB16_SMF
#define this NULL
#endif


/*
  openmidioutput is called at the start of Bochs. It should
  prepare the MIDI hardware for receiving midi commands.
   
  It is only called if the selected midimode is 1.
   
  Please note: the filenames of the SB16-options haven't been checked 
               yet, so all could be invalid values, including NULL.

  openmidioutput() will always be called before openwaveoutput(), and
  closemidioutput() will always be called before closewaveoutput(), but
  not in all cases will both functions be called.

  */

void bx_sb16_c::openmidioutput()
{
  if ( (bx_options.sb16.midifile == NULL) ||
       (strlen(bx_options.sb16.midifile) < 1) )
    {
      bx_options.sb16.midimode = 0;
      return;
    }

  MIDIDATA = fopen(bx_options.sb16.midifile,"w");
}


/* sendmidicommand should send the given midi command to the midi
   hardware.

 - delta is the number of delta ticks that have passed since the last
   command has been issued. It is always zero for the first command.
 - command is the midi command byte (sometimes called status byte), 
   in the usual range of 0x80..0xff
 - length is the number of data bytes that pertain to the command
 - data[] is the array of these data bytes. Note, it might be NULL if
   length==0.
 */ 

void bx_sb16_c::sendmidicommand(int delta, int command, int length, Bit8u data[])
{
  //  writedeltatime(delta);
  fputc(command, MIDIDATA);
  fwrite(data, 1, length, MIDIDATA);
  fflush(MIDIDATA);       // to start playing immediately
}

/* closemidioutput is called before shutting down Bochs, or if the
   emulator gets the stop_output command. After this, no more output
   will be necessary until maybe openmidioutput is called again.
   */

void bx_sb16_c::closemidioutput()
{
  fclose(MIDIDATA);
}

/* openwaveoutput should prepare the wave devices for output. It is
   only called if the selected wavemode is 1.
   */

void bx_sb16_c::openwaveoutput()
{
  if ( (bx_options.sb16.wavefile == NULL) ||
       (strlen(bx_options.sb16.wavefile) < 1) )
    {
      bx_options.sb16.wavemode = 0;
      return;
    }

  WAVEDATA = fopen(bx_options.sb16.wavefile, "w");
}

/* sendwavepacket will (in the future) receive complete packets of
   wave data. At the moment, the DMA controller can only transfer single
   bytes, so that this is not possible yet.
   At the moment it is called when a DMA transfer is complete.
   */
void bx_sb16_c::sendwavepacket()
{
  fflush(WAVEDATA);
}

/* sendwavesample gets one sample and should send it to the wave device.
   Note that this sample can be anything from 1..4 bytes, as given by
   the variable DSP.currentsamplebytes
   Note also that it is not necessarily (but probably in 99.9% of the 
   cases) part of a DMA transfer. This would be true if DSP.dmamode == 0.
   The sample is contained in the least significant bits.
   */

void bx_sb16_c::sendwavesample(Bit32u sample)
{
  fwrite(&sample, 1, DSP.currentsamplebytes, WAVEDATA );
}

/* closewaveoutput should close the wave device. No further wave
   output is necessary, at least until openwaveoutput is called again.
   */
void bx_sb16_c::closewaveoutput()
{
  fclose(WAVEDATA);
}

